from .constants import *

class ArabicPhonemizer:
    def __init__(self,
                 separator:str = ""):
        self.separator = separator

    def phonemize(self,
                  text: str) -> str:
        """
        Phonemizes a string of text.
        args:
            text (str): The text to be phonemized.
        returns:
            str: A string of phonemes corresponding to the input text.
        """
        text = self.handle_special_cases(text)
        phonemized = ""
        for char in text:
            phonemized += self._char_to_phoneme(char)
            if self.separator:
                phonemized += self.separator
        return phonemized

    def handle_special_cases(self, text: str) -> str:
        """
        Handles special cases in the input text.
        args:
            text (str): The input text.
        returns:
            str: The modified text with special cases handled.
        """
        text = self._handle_special_words(text)
        text = self._handle_alf_wasl_cases(text)
        text = self._handle_ta2_marboota_cases(text)
        text = self._handle_alf_lam_cases(text)
        return text

    def _handle_alf_lam_cases(self,
                             text: str) -> str:
        """
        Handles special cases (shamsia or qamaria) related to the letters "ال" in the input text.
        args:
            text (str): The input text.
        returns:
            str: The modified text with special cases handled.
        """
        words = text.split()
        handled_words = []
        for word in words:
            list_word = list(word)
            undiacritized_word = self._remove_diacritics(word)
            if undiacritized_word[0:4] in ["وبال","فبال","فكال","وكال","وفال","فوال"] and undiacritized_word[4] in SHAMSI_LETTERS+QAMARI_LETTERS:
                index = list_word.index(undiacritized_word[4])
                alf_index = list_word.index("ا")
                for idx, char in enumerate(list_word[alf_index:index]):
                    if char in TASHKEEL:
                        list_word.remove(char)
                list_word.remove("ا")
                lam_index = list_word.index("ل")
                if undiacritized_word[4] in SHAMSI_LETTERS:
                    list_word.remove("ل")
                    if list_word[lam_index] != "ّ": #force Shadda after lam shamsia
                        list_word.insert(lam_index+1,"ّ")
                elif undiacritized_word[4] in QAMARI_LETTERS:
                    list_word.insert(lam_index+1,"ْ")
                    
            elif undiacritized_word[0:3] in ["بال","كال","فال","وال","أال"] and undiacritized_word[3] in SHAMSI_LETTERS+QAMARI_LETTERS:
                index = list_word.index(undiacritized_word[3])
                alf_index = list_word.index("ا")
                for idx, char in enumerate(list_word[alf_index:index]):
                    if char in TASHKEEL:
                        list_word.remove(char)
                list_word.remove("ا")
                lam_index = list_word.index("ل")
                if undiacritized_word[3] in SHAMSI_LETTERS:
                    list_word.remove("ل")
                    if list_word[lam_index] != "ّ": #force Shadda after lam shamsia
                        list_word.insert(lam_index+1,"ّ")
                elif undiacritized_word[3] in QAMARI_LETTERS:
                    list_word.insert(lam_index+1,"ْ")

            elif undiacritized_word[0:2] == "ال" and undiacritized_word[2] in SHAMSI_LETTERS+QAMARI_LETTERS:
                index = list_word.index(undiacritized_word[2])
                for idx, char in enumerate(list_word[:index]):
                    if char in TASHKEEL:
                        list_word.remove(char)
                lam_index = list_word.index("ل")
                if undiacritized_word[2] in SHAMSI_LETTERS:
                    list_word.remove("ل")
                    if list_word[lam_index+1] != "ّ": #force Shadda after lam shamsia
                        list_word.insert(lam_index+1,"ّ")
                elif undiacritized_word[2] in QAMARI_LETTERS:
                    list_word.insert(lam_index+1,"ْ")
            elif undiacritized_word[0:2] == "لل" and undiacritized_word[2] in SHAMSI_LETTERS+QAMARI_LETTERS:
                index = list_word.index(undiacritized_word[2])
                lam_index_1 = list_word.index("ل")
                lam_index_2 = list_word.index("ل", lam_index_1+1)
                for idx, char in enumerate(list_word[lam_index_2:index]):
                    if char in TASHKEEL:
                        list_word.remove(char)
                lam_index_1 = list_word.index("ل")
                lam_index_2 = list_word.index("ل", lam_index_1+1)
                if undiacritized_word[2] in SHAMSI_LETTERS:
                    list_word.pop(lam_index_2)
                    if list_word[lam_index_2+1] != "ّ": #force Shadda after lam shamsia
                        list_word.insert(lam_index_2+1,"ّ")
                elif undiacritized_word[2] in QAMARI_LETTERS:
                    list_word.insert(lam_index_2+1,"ْ")
            handled_words.append("".join(list_word))
        return " ".join(handled_words)

    def _handle_alf_wasl_cases(self,
                              text:str) -> str:
        """
        Handles the case of "ا" at the beginning of a word.
        args:
            text (str): The input text.
        returns:
            str: The modified text with the case handled.
        """
        chars_list = list(text)
        matches = alf_wasl_pattern.finditer(text)
        bos_matches = alf_lam_bos.finditer(text)

        for bos_match in bos_matches:
            start = bos_match.start(2)
            end = bos_match.end(2)
            word = list(self._handle_alf_lam_cases(bos_match.group(2)))
            word[0] = "أَ"
            for i in range(start, end):
                chars_list[i] = ""
            chars_list[start] = "".join(word)

        for match in matches:
            idx = match.start()
            if self._starts_with_alf_lam(match.group()): 
                pass
            elif chars_list[idx+1] == "َ":
                chars_list[idx] = "أ"
            elif chars_list[idx+1] == "ُ":
                chars_list[idx] = "أ"
            elif chars_list[idx+1] == "ِ":
                chars_list[idx] = "إ"
            elif chars_list[idx+1] in ARABIC_LETTERS:
                chars_list[idx] = "إِ"
            else:
                pass
        
        return "".join(chars_list)

    def _handle_ta2_marboota_cases(self,
                                   text: str) -> str:
        """
        Handles the case of "ة" at the ending of a word.
        args:
            text (str): The input text.
        returns:
            str: The modified text with the case handled.
        """
        words = text.split()
        chars_list = list(text)
        matches = ta2_marbootah_sakenah_pattern.finditer(text)
        for match in matches:
            if self._starts_with_alf_lam(match.group()):
                chars_list[match.end()-2:match.end()] = "هْ"
            else:
                if words_pattern.match(text[match.end()+1:]):
                    chars_list[match.end()-2:match.end()] = "تْ"
                else:
                    chars_list[match.end()-2:match.end()] = "هْ"
        return "".join(chars_list)
    
    def _handle_special_words(self,
                              text:str) -> str:
        """
        Handles words with special pronounciation. See constants.py for more info.
        args:
            text (str): The input text.
        returns:
            str: The modified text with special cases handled.
        """
        words = text.split()
        handled_words = []
        for word in words:
            undiacritized_word = self._remove_diacritics(word)
            if undiacritized_word in SPECIAL_WORDS_MAP:
                handled_words.append(SPECIAL_WORDS_MAP[undiacritized_word])
            else:
                handled_words.append(word)
        return " ".join(handled_words)

    def _handle_undiacritizable_alf(self,
                                     text:str) -> str:
        """
        Handles the case where the letter "ا" shouldn't be diacritized.
        args:
            text (str): The input text.
        returns:
            str: The modified text with the special case handled.
        """
        word_list = list(text)
        for i in range(len(word_list)):
            if word_list[i] == "ا":
                #Todo
                pass

    def _char_to_phoneme(self, char: str) -> str:
        """
        Converts a character to its corresponding phoneme representation.
        args:
            char (str): The character to be converted.
        returns:
            str: The phoneme representation of the character.
        """
        if char in ARABIC2BUCKWALTER:
            return ARABIC2BUCKWALTER[char]
        else:
            return char

    def _remove_diacritics(self,
                           text: str) -> str:
        """
        Removes diacritics from the input text.
        args:
            text (str): The input text.
        returns:
            str: The text with diacritics removed.
        """
        for char in TASHKEEL:
            text = text.replace(char, '')
        return text
    
    def _starts_with_alf_lam(self,
                             text:str) -> bool:
        """
        Checks if the input text starts with the letter "ال".
        args:
            text (str): The input text.
        returns:
            bool: True if the text starts with "ال", False otherwise.
        """
        patterns = [alf_lam_pattern, char_alf_lam_pattern, char_char_alf_lam_pattern]
        if any(pattern.match(text) for pattern in patterns):
            return True
        return False


import pytest
from arabic_phonemizer import ArabicPhonemizer  # Replace with the actual import path

@pytest.fixture
def phonemizer():
    return ArabicPhonemizer()

@pytest.fixture
def texts():
    data = {'ذَهَبَ الْوَلَدُ إِلَى الْمَدْرَسَةِ الْجَدِيدَةْ.': 'ذَهَبَ الْوَلَدُ إِلَى الْمَدْرَسَةِ الْجَدِيدَهْ.',
            'بِالْكِتَابِ تَرْتَقِي الْأُمَمُ.': 'بِلْكِتَابِ تَرْتَقِي الْأُمَمُ.',
            'فِي السَّمَاءِ تَطِيرُ الطُّيُورُ بِحُرِّيَّةْ.': 'فِي اسَّمَاءِ تَطِيرُ اطُّيُورُ بِحُرِّيَّهْ.',
            'لِلْعِلْمِ قِيمَةٌ كَبِيرَةٌ فِي الْحَيَاةِ.': 'لِلْعِلْمِ قِيمَةٌ كَبِيرَةٌ فِي الْحَيَاةِ.',
            ' الْقَمَرُ مُنِيرٌ فِي اللَّيْلِ الصَّافِي!': 'أَلْقَمَرُ مُنِيرٌ فِي الَّيْلِ اصَّافِي!',
            'اَلشَّمْسُ سَاطِعَةْ، مُنِيرَةٌ فِي النَّهَارِ الْجَمِيلِ.': 'أَشَّمْسُ سَاطِعَهْ، مُنِيرَةٌ فِي انَّهَارِ الْجَمِيلِ.',
            'اِسْتَمَعْتُ إِلَى الْخَبَرِ السَّارِّ بِاهْتِمَامٍ.': 'إِسْتَمَعْتُ إِلَى الْخَبَرِ اسَّارِّ بِاهْتِمَامٍ.',
            'أَكَلْتُ الْفَاكِهَةَ اللَّذِيذَةَ بَعْدَ الْغَدَاءِ.': 'أَكَلْتُ الْفَاكِهَةَ الَّذِيذَةَ بَعْدَ الْغَدَاءِ.',
            'اِشْتَرَيْتُ الْقَلَمَ وَالدَّفْتَرَ مِنَ الْمَكْتَبَةِ.': 'إِشْتَرَيْتُ الْقَلَمَ وَدَّّفْتَرَ مِنَ الْمَكْتَبَةِ.',
            'الْكِتَابُ، الْكِتَابُ هُوَ خَيْرُ صَدِيقٍ.': 'أَلْكِتَابُ، أَلْكِتَابُ هُوَ خَيْرُ صَدِيقٍ.',
            'اَلصَّدِيقُ الْوَفِيُّ كَنْزٌ لَا يُقَدَّرُ بِثَمَنٍ.': 'أَصَّدِيقُ الْوَفِيُّ كَنْزٌ لَا يُقَدَّرُ بِثَمَنٍ.',
            'بِالْعَمَلِ الْجَادِّ نَحْقِقُ الْأَهْدَافَ الْمَرْجُوَّةَ.': 'بِلْعَمَلِ الْجَادِّ نَحْقِقُ الْأَهْدَافَ الْمَرْجُوَّةَ.',
            'لِلْأَطْفَالِ حَقٌّ فِي التَّعْلِيمِ وَاللَّعِبِ.': 'لِلْأَطْفَالِ حَقٌّ فِي اتَّعْلِيمِ وَلَّّعِبِ.',
            'اِسْتَعَانَ الْمُهَنْدِسُ بِالْأَدَوَاتِ الْحَدِيثَةِ.': 'إِسْتَعَانَ الْمُهَنْدِسُ بِلْأَدَوَاتِ الْحَدِيثَةِ.',
            'اِقْتَرَبَ الْعِيدُ وَفَرِحَ الْجَمِيعُ بِهِ.': 'إِقْتَرَبَ الْعِيدُ وَفَرِحَ الْجَمِيعُ بِهِ.',
            'هذا ذلك طه هؤلاء أولئك':'هَاذَا ذَالِكَ طَاهَا هَاؤُلَاءِ أولَائِكَ'}
    return data

def test_special_cases(texts): 
    phonemizer = ArabicPhonemizer()
    for text, expected in texts.items():
        phonemized = phonemizer.handle_special_cases(text)
        assert phonemized == expected


import re 

ARABIC2BUCKWALTER = { #mapping from Arabic script to Buckwalter
	u'\u0628': u'b' , u'\u0630': u'*' , u'\u0637': u'T' , u'\u0645': u'm',
	u'\u062a': u't' , u'\u0631': u'r' , u'\u0638': u'Z' , u'\u0646': u'n',
	u'\u062b': u'^' , u'\u0632': u'z' , u'\u0639': u'E' , u'\u0647': u'h',
	u'\u062c': u'j' , u'\u0633': u's' , u'\u063a': u'g' , u'\u062d': u'H',
	u'\u0642': u'q' , u'\u0641': u'f' , u'\u062e': u'x' , u'\u0635': u'S',
	u'\u0634': u'$' , u'\u062f': u'd' , u'\u0636': u'D' , u'\u0643': u'k',
	u'\u0623': u'>' , u'\u0621': u'\'', u'\u0626': u'}' , u'\u0624': u'&',
	u'\u0625': u'<' , u'\u0622': u'|' , u'\u0627': u'A' , u'\u0649': u'Y',
	u'\u0629': u'p' , u'\u064a': u'y' , u'\u0644': u'l' , u'\u0648': u'w',
	u'\u064b': u'F' , u'\u064c': u'N' , u'\u064d': u'K' , u'\u064e': u'a',
	u'\u064f': u'u' , u'\u0650': u'i' , u'\u0651': u'~' , u'\u0652': u'o'
}

SPECIAL_WORDS_MAP = { #Regularly being updated
    "هذا": "هَاذَا",
    "هذه": "هَاذِهِ",
    "هذان": "هَاذَانِ",
    "هؤلاء": "هَاؤُلَاءِ",
    "ذلك": "ذَالِكَ",
    "أولئك": "أولَائِكَ",
    "طه": "طَاهَا",
    "لكن": "لَاكِنْ",
    "لكنه": "لَاكِنَّهُ",
    "لكنك": "لَاكِنَّكَ",
    "لكنكم": "لَاكِنَّكُمْ",
    "لكنهم": "لَاكِنَّهُمْ",
    "لكنهما": "لَاكِنَّهُمَا",
    "لكننا": "لَاكِنَّنَا",
    "الذي": "اللَّذِي",
    "التي": "اللَّتِي",
    "الذين": "اللَّذِينَ",
}

ARABIC_LETTERS = "ابتثجحخدذرزسشصضطظعغفقكلمنهويىءةؤئإأآ"
SHAMSI_LETTERS = "تثدذرزسشصضطظلن"
QAMARI_LETTERS = "اأإآبجحخعغفقكمهوي"
TASHKEEL = "ًٌٍَُِّْ"
PUNCTUATIONS = "".join(["،",",","؛"",",":",",","؟","?","\\-",",","_",",",".",";","\"","\'"])

alf_wasl_pattern = re.compile(f"\\b(?<![{TASHKEEL}])ا[{ARABIC_LETTERS+TASHKEEL}]+\\b",re.UNICODE)
# alf_mad_pattern = re.compile(f"{wb}([{ARABIC_LETTERS+TASHKEEL}]+)(ا[{TASHKEEL}]*)([{ARABIC_LETTERS+TASHKEEL}]+){wb}",re.UNICODE)
ta2_marbootah_sakenah_pattern = re.compile(f"\\b[{ARABIC_LETTERS+TASHKEEL}]+ةْ",re.UNICODE)
words_pattern = re.compile(f"\\b(?<![{PUNCTUATIONS}])[{ARABIC_LETTERS+TASHKEEL}]+\\b",re.UNICODE)
char_alf_lam_pattern = re.compile(f"\\b[كفوب][{TASHKEEL}]*ا[{TASHKEEL}]*ل[{TASHKEEL}]*[{ARABIC_LETTERS+TASHKEEL}]+\\b",re.UNICODE)
char_char_alf_lam_pattern = re.compile(f"\\b[وف][{TASHKEEL}]*[كفوب][{TASHKEEL}]*ا[{TASHKEEL}]*ل[{TASHKEEL}]*[{ARABIC_LETTERS+TASHKEEL}]+\\b",re.UNICODE)
alf_lam_pattern = re.compile(f"\\bا[{TASHKEEL}]*ل[{TASHKEEL}]*[{ARABIC_LETTERS+TASHKEEL}]+\\b",re.UNICODE)
alf_lam_bos = re.compile(f"(^|[{PUNCTUATIONS}][\\s]*)(ا[{TASHKEEL}]*ل[{TASHKEEL}]*[{ARABIC_LETTERS+TASHKEEL}]+(\\b))",re.UNICODE)
arabic_letters_pattern = re.compile(f"[{ARABIC_LETTERS}]",re.UNICODE)
all_arabic_pattern = re.compile(f"[{ARABIC_LETTERS+TASHKEEL+PUNCTUATIONS} ]+",re.UNICODE)
punctuations_pattern = re.compile(f"[{PUNCTUATIONS}]",re.UNICODE)
sentence_segments_pattern = re.compile(f"[^{PUNCTUATIONS}]+[{PUNCTUATIONS}]+",re.UNICODE)
no_punc_segments_pattern = re.compile(f"[^{PUNCTUATIONS}]+$",re.UNICODE)